##############################################################################
# pymbar: A Python Library for MBAR
#
# Copyright 2016-2017 University of Colorado Boulder
# Copyright 2010-2017 Memorial Sloan-Kettering Cancer Center
# Portions of this software are Copyright 2010-2016 University of Virginia
#
# Authors: Michael Shirts, John Chodera
# Contributors: Kyle Beauchamp, Levi Naden
#
# pymbar is free software: you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# MIT License for more details.
#
# You should have received a copy of the MIT License along with pymbar.
##############################################################################

##############################################################################
# imports
##############################################################################

from itertools import zip_longest
import warnings
import numpy as np
import numexpr


##############################################################################
# functions / classes
##############################################################################


class TypeCastPerformanceWarning(RuntimeWarning):
    pass


def kln_to_kn(kln, N_k=None, cleanup=False):
    """Convert KxKxN_max array to KxN max array

    Parameters
    ----------
    u_kln : np.ndarray, float, shape=(KxLxN_max)
    N_k : np.array, optional
        the N_k matrix from the previous formatting form
    cleanup : bool, optional
        optional command to clean up, since u_kln can get very large

    Returns
    -------
    u_kn : np.ndarray, float, shape=(LxN)
    """

    # rewrite into kn shape
    K, L, N_max = np.shape(kln)

    if N_k is None:
        # We assume that all N_k are N_max.
        # Not really an easier way to do this without being given the answer.
        N_k = N_max * np.ones([L], dtype=np.int64)
    N = np.sum(N_k)

    kn = np.zeros([L, N], dtype=np.float64)
    i = 0
    for k in range(K):  # loop through the old K; some might be zero
        for ik in range(N_k[k]):
            kn[:, i] = kln[k, :, ik]
            i += 1
    if cleanup:
        del kln  # very big, let's explicitly delete

    return kn


def kn_to_n(kn, N_k=None, cleanup=False):
    """Convert KxN_max array to N array

    Parameters
    ----------
    u_kn : np.ndarray, float, shape=(KxN_max)
    N_k : np.array, optional
        the N_k matrix from the previous formatting form
    cleanup : bool, optional
        optional command to clean up, since u_kln can get very large

    Returns
    -------
    u_n : np.ndarray, float, shape=(N)
    """

    # print "warning: KxN arrays deprecated; converting into new preferred N shape"
    # rewrite into kn shape

    # rewrite into kn shape
    K, N_max = np.shape(kn)

    if N_k is None:
        # We assume that all N_k are N_max.
        # Not really an easier way to do this without being given the answer.
        N_k = N_max * np.ones([K], dtype=np.int64)
    N = np.sum(N_k)

    n = np.zeros([N], dtype=np.float64)
    i = 0
    for k in range(K):  # loop through the old K; some might be zero
        for ik in range(N_k[k]):
            n[i] = kn[k, ik]
            i += 1
    if cleanup:
        del kn  # very big, let's explicitly delete
    return n


def ensure_type(
    val,
    dtype,
    ndim,
    name,
    length=None,
    can_be_none=False,
    shape=None,
    warn_on_cast=True,
    add_newaxis_on_deficient_ndim=False,
):
    """Typecheck the size, shape and dtype of a numpy array, with optional
    casting.

    Parameters
    ----------
    val : {np.ndaraay, None}
        The array to check
    dtype : {nd.dtype, str}
        The dtype you'd like the array to have
    ndim : int
        The number of dimensions you'd like the array to have
    name : str
        name of the array. This is used when throwing exceptions, so that
        we can describe to the user which array is messed up.
    length : int, optional
        How long should the array be?
    can_be_none : bool
        Is ``val == None`` acceptable?
    shape : tuple, optional
        What should be shape of the array be? If the provided tuple has
        Nones in it, those will be semantically interpreted as matching
        any length in that dimension. So, for example, using the shape
        spec ``(None, None, 3)`` will ensure that the last dimension is of
        length three without constraining the first two dimensions
    warn_on_cast : bool, default=True
        Raise a warning when the dtypes don't match and a cast is done.
    add_newaxis_on_deficient_ndim : bool, default=True
        Add a new axis to the beginning of the array if the number of
        dimensions is deficient by one compared to your specification. For
        instance, if you're trying to get out an array of ``ndim == 3``,
        but the user provides an array of ``shape == (10, 10)``, a new axis will
        be created with length 1 in front, so that the return value is of
        shape ``(1, 10, 10)``.

    Notes
    -----
    The returned value will always be C-contiguous.

    Returns
    -------
    typechecked_val : np.ndarray, None
        If `val=None` and `can_be_none=True`, then this will return None.
        Otherwise, it will return val (or a copy of val). If the dtype wasn't right,
        it'll be casted to the right shape. If the array was not C-contiguous, it'll
        be copied as well.

    """
    if can_be_none and val is None:
        return None

    if not isinstance(val, np.ndarray):
        # special case: if the user is looking for a 1d array, and
        # they request newaxis upconversion, and provided a scalar
        # then we should reshape the scalar to be a 1d length-1 array
        if add_newaxis_on_deficient_ndim and ndim == 1 and np.isscalar(val):
            val = np.array([val])
        else:
            raise TypeError(
                ("{} must be numpy array. " " You supplied type {}".format(name, type(val)))
            )

    if warn_on_cast and val.dtype != dtype:
        warnings.warn(
            "Casting {} dtype={} to {} ".format(name, val.dtype, dtype), TypeCastPerformanceWarning
        )

    if not val.ndim == ndim:
        if add_newaxis_on_deficient_ndim and val.ndim + 1 == ndim:
            val = val[np.newaxis, ...]
        else:
            raise ValueError(
                ("{} must be ndim {}. " "You supplied {}".format(name, ndim, val.ndim))
            )

    val = np.ascontiguousarray(val, dtype=dtype)

    if length is not None and len(val) != length:
        raise ValueError(
            ("{} must be length {}. " "You supplied {}.".format(name, length, len(val)))
        )

    if shape is not None:
        # the shape specified given by the user can look like (None, None 3)
        # which indicates that ANY length is accepted in dimension 0 or
        # dimension 1
        sentinel = object()
        error = ValueError(
            (
                "{} must be shape {}. You supplied  "
                "{}".format(name, str(shape).replace("None", "Any"), val.shape)
            )
        )
        for a, b in zip_longest(val.shape, shape, fillvalue=sentinel):
            if a is sentinel or b is sentinel:
                # if the sentinel was reached, it means that the ndim didn't
                # match or something. this really shouldn't happen
                raise error
            if b is None:
                # if the user's shape spec has a None in it, it matches anything
                continue
            if a != b:
                # check for equality
                raise error

    return val


def _logsum(a_n):
    """Compute the log of a sum of exponentiated terms exp(a_n) in a numerically-stable manner.

    NOTE: this function has been deprecated in favor of logsumexp.

    Parameters
    ----------
    a_n : np.ndarray, shape=(n_samples)
        a_n[n] is the nth exponential argument

    Returns
    -------
    a_n : np.ndarray, shape=(n_samples)
        a_n[n] is the nth exponential argument

    Notes
    -----

    _logsum a_n = max_arg + \\log \\sum_{n=1}^N \\exp[a_n - max_arg]

    where max_arg = max_n a_n.  This is mathematically (but not numerically) equivalent to

    _logsum a_n = \\log \\sum_{n=1}^N \\exp[a_n]


    Example
    -------
    >>> a_n = np.array([0.0, 1.0, 1.2], np.float64)
    >>> print('{:.3e}'.format(_logsum(a_n)))
    1.951e+00
    """

    # Compute the maximum argument.
    max_log_term = np.max(a_n)

    # Compute the reduced terms.
    terms = np.exp(a_n - max_log_term)

    # Compute the log sum.
    log_sum = np.log(np.sum(terms)) + max_log_term

    return log_sum


def logsumexp(a, axis=None, b=None, use_numexpr=True):
    """Compute the log of the sum of exponentials of input elements.

    Parameters
    ----------
    a : array_like
        Input array.
    axis : None or int, optional, default=None
        Axis or axes over which the sum is taken. By default `axis` is None,
        and all elements are summed.
    b : array-like, optional
        Scaling factor for exp(`a`) must be of the same shape as `a` or
        broadcastable to `a`.
    use_numexpr : bool, optional, default=True
        If True, use the numexpr library to speed up the calculation, which
        can give a 2-4X speedup when working with large arrays.

    Returns
    -------
    res : ndarray
        The result, ``log(sum(exp(a)))`` calculated in a numerically
        more stable way. If `b` is given then ``log(sum(b*exp(a)))``
        is returned.

    See Also
    --------
    numpy.logaddexp, numpy.logaddexp2, scipy.special.logsumexp

    Notes
    -----
    This is based on ``scipy.special.logsumexp`` but with optional numexpr
    support for improved performance.
    """

    a = np.asarray(a)

    a_max = np.amax(a, axis=axis, keepdims=True)

    if a_max.ndim > 0:
        a_max[~np.isfinite(a_max)] = 0
    elif not np.isfinite(a_max):
        a_max = 0

    if b is not None:
        b = np.asarray(b)
        if use_numexpr:
            out = np.log(numexpr.evaluate("b * exp(a - a_max)").sum(axis))
        else:
            out = np.log(np.sum(b * np.exp(a - a_max), axis=axis))
    else:
        if use_numexpr:
            out = np.log(numexpr.evaluate("exp(a - a_max)").sum(axis))
        else:
            out = np.log(np.sum(np.exp(a - a_max), axis=axis))

    a_max = np.squeeze(a_max, axis=axis)
    out += a_max

    return out


def check_w_normalized(W, N_k, tolerance=1.0e-4):
    """Check the weight matrix W is properly normalized. The sum over N should be 1, and the sum over k by N_k should aslo be 1

    Parameters
    ----------
    W : np.ndarray, shape=(N, K), dtype='float'
        The normalized weight matrix for snapshots and states.
        W[n, k] is the weight of snapshot n in state k.
    N_k : np.ndarray, shape=(K), dtype='int'
        N_k[k] is the number of samples from state k.
    tolerance : float, optional, default=1.0e-4
        Tolerance for checking equality of sums

    Returns
    -------
    None : NoneType
        Returns a None object if test passes

    Raises
    ------
    ParameterError
        Appropriate message if W is not normalized within tolerance.
    """

    N, K = W.shape

    column_sums = np.sum(W, axis=0)
    badcolumns = np.abs(column_sums - 1) > tolerance
    if np.any(badcolumns):
        which_badcolumns = np.arange(K)[badcolumns]
        firstbad = which_badcolumns[0]
        raise ParameterError(
            (
                "Warning: Should have \\sum_n W_nk = 1. "
                f"Actual column sum for state {firstbad:d} was {column_sums[firstbad]:f}. "
                f"{np.sum(badcolumns):d} other columns have similar problems. \n"
                "This generally indicates the free energies are not converged."
            )
        )

    row_sums = np.sum(W * N_k, axis=1)
    badrows = np.abs(row_sums - 1) > tolerance
    if np.any(badrows):
        which_badrows = np.arange(N)[badrows]
        firstbad = which_badrows[0]
        raise ParameterError(
            (
                "Warning: Should have \\sum__k N_k W_nk = 1. "
                f"Actual row sum for state {firstbad:d} was {row_sums[firstbad]:f}. "
                f"{np.sum(badrows):d} other columns have similar problems. \n"
                "This generally indicates the free energies are not converged."
            )
        )
    return


# ============================================================================================
# Exception classes
# =============================================================================================


class ParameterError(Exception):
    """
    An error in the input parameters has been detected.
    """


class ConvergenceError(Exception):
    """
    Convergence could not be achieved.
    """


class BoundsError(Exception):
    """
    Could not determine bounds on free energy
    """


class DataError(Exception):
    """
    Data is inconsistent.
    """



"""
Example illustrating the application of MBAR to compute a 1D free energy profile
from a series of force-clamp single-molecule experiments.

REFERENCE

    Woodside MT, Behnke-Parks WL, Larizadeh K, Travers K, Herschlag D, and Block SM.
    Nanomechanical measurements of the sequence-dependent folding landscapes of single
    nucleic acid hairpins. PNAS 103:6190, 2006.

"""

# =============================================================================================
# IMPORTS
# =============================================================================================
import subprocess
import time
from pathlib import Path

import numpy as np
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt

import pymbar  # multistate Bennett acceptance ratio analysis (provided by pymbar)
from pymbar import timeseries, FES  # timeseries analysis (provided by pymbar)

# =============================================================================================
# PARAMETERS
# =============================================================================================

prefix = "20R55_4T"  # for paper
# prefix = '10R50_4T'
# prefix = '25R50_4T'
# prefix = '30R50_4T'
directory = Path("processed-data")
temperature = 296.15  # temperature (in K)
nbins = 50  # number of bins for 1D FES
output_directory = Path("output")
plot_directory = Path("plots")

# =============================================================================================
# CONSTANTS
# =============================================================================================

kB = 1.381e-23  # Boltzmann constant (in J/K)
pN_nm_to_kT = (1.0e-9) * (1.0e-12) / (kB * temperature)  # conversion from nM pN to units of kT

# =============================================================================================
# SUBROUTINES
# =============================================================================================


def construct_nonuniform_bins(x_n, nbins):
    """Construct histogram using bins of unequal size to ensure approximately equal population in each bin.

    Parameters
    ----------
    x_n : 1D array of float
        x_n[n] is data point n

    Returns
    -------
    bin_left_boundary_i : 1D array of floats
        data in bin i will satisfy bin_left_boundary_i[i] <= x < bin_left_boundary_i[i+1]
    bin_center_i : 1D array of floats
        bin_center_i[i] is the center of bin i
    bin_width_i : 1D array of floats
        bin_width_i[i] is the width of bin i
    bin_n : 1D array of int32
        bin_n[n] is the bin index (in range(nbins)) of x_n[n]
    """

    # Determine number of samples.
    N = x_n.size

    # Get indices of elements of x_n sorted in order.
    sorted_indices = x_n.argsort()

    # Allocate storage for results.
    bin_left_boundary_i = np.zeros([nbins + 1])
    bin_center_i = np.zeros([nbins])
    bin_width_i = np.zeros([nbins])
    bin_n = np.zeros([N])

    # Determine sampled range, adding a little bit to the rightmost range to ensure no samples escape the range.
    x_min = x_n.min()
    x_max = x_n.max()
    x_max += (x_max - x_min) * 1.0e-5

    # Determine bin boundaries and bin assignments.
    for bin_index in range(nbins):
        # indices of first and last data points in this span
        first_index = int(float(N) / float(nbins) * float(bin_index))
        last_index = int(float(N) / float(nbins) * float(bin_index + 1))

        # store left bin boundary
        bin_left_boundary_i[bin_index] = x_n[sorted_indices[first_index]]

        # store assignments
        bin_n[sorted_indices[first_index:last_index]] = bin_index

    # set rightmost boundary
    bin_left_boundary_i[nbins] = x_max

    # Determine bin centers and widths
    for bin_index in range(nbins):
        bin_center_i[bin_index] = (
            bin_left_boundary_i[bin_index] + bin_left_boundary_i[bin_index + 1]
        ) / 2.0
        bin_width_i[bin_index] = (
            bin_left_boundary_i[bin_index + 1] - bin_left_boundary_i[bin_index]
        )

    return bin_left_boundary_i, bin_center_i, bin_width_i, bin_n


# =============================================================================================
# MAIN
# =============================================================================================


def main():
    # read biasing forces for different trajectories
    filename = directory / f"{prefix}.forces"
    with open(filename) as infile:
        elements = infile.readline().split()
        K = len(elements)  # number of biasing forces
        # biasing_force_k[k] is the constant external biasing force used to collect trajectory k (in pN)
        biasing_force_k = np.zeros([K])
        for k in range(K):
            biasing_force_k[k] = float(elements[k])
    print("biasing forces (in pN) = ", biasing_force_k)

    # Determine maximum number of snapshots in all trajectories.
    filename = directory / f"{prefix}.trajectories"
    T_max = 0
    with open(filename) as f:
        for line in f:
            T_max += 1

    # Allocate storage for original (correlated) trajectories
    T_k = np.zeros([K], int)  # T_k[k] is the number of snapshots from umbrella simulation k`
    # x_kt[k,t] is the position of snapshot t from trajectory k (in nm)
    x_kt = np.zeros([K, T_max])

    # Read the trajectories.
    filename = directory / f"{prefix}.trajectories"
    print(f"Reading {filename}...")
    with open(filename) as infile:
        for line in infile:
            elements = line.split()
            for k in range(K):
                t = T_k[k]
                x_kt[k, t] = float(elements[k])
                T_k[k] += 1

    # Create a list of indices of all configurations in kt-indexing.
    mask_kt = np.zeros([K, T_max], dtype=bool)
    for k in range(K):
        mask_kt[k, 0 : T_k[k]] = True
    # Create a list from this mask.
    all_data_indices = np.where(mask_kt)

    # Construct equal-frequency extension bins
    print("binning data...")
    bin_kt = np.zeros([K, T_max], int)
    bin_left_boundary_i, bin_center_i, bin_width_i, bin_assignments = construct_nonuniform_bins(
        x_kt[all_data_indices], nbins
    )
    bin_kt[all_data_indices] = bin_assignments

    # Compute correlation times.
    N_max = 0
    g_k = np.zeros([K])
    for k in range(K):
        # Compute statistical inefficiency for extension timeseries
        g = timeseries.statistical_inefficiency(x_kt[k, 0 : T_k[k]], x_kt[k, 0 : T_k[k]])
        # store statistical inefficiency
        g_k[k] = g
        print(
            f"timeseries {k + 1:d} : g = {g:.1f}, {int(np.floor(T_k[k] / g)):.0f} "
            f"uncorrelated samples (of {T_k[k]:d} total samples)"
        )
        N_max = max(N_max, int(np.ceil(T_k[k] / g)) + 1)

    # Subsample trajectory position data.
    x_kn = np.zeros([K, N_max])
    bin_kn = np.zeros([K, N_max])
    N_k = np.zeros([K], int)
    for k in range(K):
        # Compute correlation times for potential energy and chi timeseries.
        indices = timeseries.subsample_correlated_data(x_kt[k, 0 : T_k[k]])
        # Store subsampled positions.
        N_k[k] = len(indices)
        x_kn[k, 0 : N_k[k]] = x_kt[k, indices]
        bin_kn[k, 0 : N_k[k]] = bin_kt[k, indices]

    # Set arbitrarynp.zeros for external biasing potential.
    x0_k = np.zeros([K])  # x position corresponding to zero of potential
    for k in range(K):
        x0_k[k] = x_kn[k, 0 : N_k[k]].mean()
    print("x0_k = ", x0_k)

    # Compute bias energies in units of kT.
    # u_kln[k,l,n] is the reduced (dimensionless) relative potential energy of
    # snapshot n from umbrella simulation k evaluated at umbrella l
    u_kln = np.zeros([K, K, N_max])
    for k in range(K):
        for l in range(K):
            # compute relative energy difference from sampled state to each other state
            # U_k(x) = F_k x
            # where F_k is external biasing force
            # (F_k pN) (x nm) (pN /
            # u_kln[k,l,0:N_k[k]] = - pN_nm_to_kT * (biasing_force_k[l] - biasing_force_k[k]) * x_kn[k,0:N_k[k]]
            u_kln[k, l, 0 : N_k[k]] = -pN_nm_to_kT * biasing_force_k[l] * (
                x_kn[k, 0 : N_k[k]] - x0_k[l]
            ) + pN_nm_to_kT * biasing_force_k[k] * (x_kn[k, 0 : N_k[k]] - x0_k[k])

    # DEBUG
    start_time = time.time()

    # Initialize MBAR.
    print("Running MBAR...")
    # TODO: change to u_kn inputs
    mbar = pymbar.MBAR(
        u_kln, N_k, verbose=True, relative_tolerance=1.0e-10, solver_protocol="robust"
    )

    # Compute unbiased energies (all biasing forces are zero).
    # u_n[n] is the reduced potential energy without umbrella restraints of snapshot n
    u_n = np.zeros([np.sum(N_k)])
    x_n = np.zeros([np.sum(N_k)])

    Nstart = 0
    for k in range(K):
        #    u_n[N_k[k]:N_k[k+1]] = - pN_nm_to_kT * (0.0 - biasing_force_k[k]) * x_kn[k,0:N_k[k]]
        u_n[Nstart : Nstart + N_k[k]] = 0.0 + pN_nm_to_kT * biasing_force_k[k] * (
            x_kn[k, 0 : N_k[k]] - x0_k[k]
        )
        x_n[Nstart : Nstart + N_k[k]] = x_kn[k, 0 : N_k[k]]
        Nstart += N_k[k]

    # Compute free energy profile in unbiased potential (in units of kT).

    print("Computing free energy profile...")
    fes = FES(u_kln, N_k, mbar_options=dict(solver_protocol="robust"))
    histogram_parameters = dict()
    # 1D array of parameters, one entry because 1D
    histogram_parameters["bin_edges"] = bin_left_boundary_i
    fes.generate_fes(u_n, x_n, histogram_parameters=histogram_parameters)
    results = fes.get_fes(
        bin_center_i, reference_point="from-lowest", uncertainty_method="analytical"
    )
    f_i = results["f_i"]
    df_i = results["df_i"]

    # compute estimate of FES including Jacobian term
    fes_i = f_i + np.log(bin_width_i)
    # Write out unbiased estimate of FES
    print("Unbiased FES (in units of kT)")
    print(f"{'bin':>8s} {'f':>8s} {'df':>8s} {'fes':>8s} {'width':>8s}")
    for i in range(nbins):
        print(
            f"{bin_center_i[i]:8.3f}",
            f"{f_i[i]:8.3f}",
            f"{df_i[i]:8.3f}",
            f"{fes_i[i]:8.3f}",
            f"{bin_width_i[i]:8.3f}",
        )

    filename = output_directory / "fes-unbiased.out"
    with open(filename, "w") as outfile:
        for i in range(nbins):
            outfile.write(f"{bin_center_i[i]:8.3f} {fes_i[i]:8.3f} {df_i[i]:8.3f}\n")

    fig_filename = plot_directory / f"fes-unbiased.pdf"
    fig, ax = plt.subplots()
    ax.errorbar(bin_center_i, fes_i, yerr=df_i, fmt="_ ", elinewidth=0.3)
    ax.set_title("Unbiased estimate of free energy profile")
    ax.set_ylabel("Free energy profile of mean force (kT)")
    ax.set_xlabel("Extension (nm)")
    fig.savefig(fig_filename)

    # DEBUG
    stop_time = time.time()
    elapsed_time = stop_time - start_time
    print(f"analysis took {elapsed_time:f} seconds")

    # compute observed and expected histograms at each state
    for l in range(K):
        # compute FES at state l
        Nstart = 0
        for k in range(K):
            u_n[Nstart : Nstart + N_k[k]] = u_kln[k, l, 0 : N_k[k]]
            Nstart += N_k[k]
        fes.generate_fes(u_n, x_n, histogram_parameters=histogram_parameters)
        results = fes.get_fes(
            bin_center_i, reference_point="from-lowest", uncertainty_method="analytical"
        )
        f_i = results["f_i"]
        df_i = results["df_i"]

        # compute estimate of FES including Jacobian term
        fes_i = f_i + np.log(bin_width_i)
        # center fes
        fes_i -= fes_i.mean()
        # compute probability distribution
        p_i = np.exp(-f_i + f_i.min())
        p_i /= p_i.sum()
        # compute observed histograms, filtering to within [x_min,x_max] range
        N_i_observed = np.zeros([nbins])
        dN_i_observed = np.zeros([nbins])
        for t in range(T_k[l]):
            bin_index = bin_kt[l, t]
            N_i_observed[bin_index] += 1
        N = N_i_observed.sum()
        # estimate uncertainties in observed counts
        for bin_index in range(nbins):
            dN_i_observed[bin_index] = np.sqrt(
                g_k[l] * N_i_observed[bin_index] * (1.0 - N_i_observed[bin_index] / float(N))
            )
        # compute expected histograms
        N_i_expected = float(N) * p_i
        # only approximate, since correlations df_i df_j are neglected
        dN_i_expected = np.sqrt(float(N) * p_i * (1.0 - p_i))
        # plot
        print(f"state {l:d} ({biasing_force_k[l]:f} pN)")
        for bin_index in range(nbins):
            print(
                f"{bin_center_i[bin_index]:8.3f}",
                f"{N_i_expected[bin_index]:10f}",
                f"{N_i_observed[bin_index]:10f} +- {dN_i_observed[bin_index]:10f}",
            )

        # Write out observed bin counts
        filename = output_directory / f"counts-observed-{l:d}.out"
        with open(filename, "w") as outfile:
            for i in range(nbins):
                outfile.write(
                    f"{bin_center_i[i]:8.3f} {N_i_observed[i]:16f} {dN_i_observed[i]:16f}\n"
                )

        # write out expected bin counts
        filename = output_directory / f"counts-expected-{l:d}.out"
        with open(filename, "w") as outfile:
            for i in range(nbins):
                outfile.write(
                    f"{bin_center_i[i]:8.3f} {N_i_expected[i]:16f} {dN_i_expected[i]:16f}\n"
                )

        # compute FES from observed counts
        indices = np.where(N_i_observed > 0)[0]
        fes_i_observed = np.zeros([nbins])
        dfes_i_observed = np.zeros([nbins])
        fes_i_observed[indices] = -np.log(N_i_observed[indices]) + np.log(bin_width_i[indices])
        fes_i_observed[indices] -= fes_i_observed[indices].mean()  # shift observed FES
        dfes_i_observed[indices] = dN_i_observed[indices] / N_i_observed[indices]
        # write out observed FES
        filename = output_directory / f"fes-observed-{l:d}.out"
        with open(filename, "w") as outfile:
            for i in indices:
                outfile.write(
                    f"{bin_center_i[i]:8.3f} {fes_i_observed[i]:8.3f} {dfes_i_observed[i]:8.3f}\n"
                )

        # Write out unbiased estimate of FES
        fes_i -= fes_i[indices].mean()  # shift to align with observed
        filename = output_directory / f"fes-expected-{l:d}.out"
        with open(filename, "w") as outfile:
            for i in range(nbins):
                outfile.write(f"{bin_center_i[i]:8.3f} {fes_i[i]:8.3f} {df_i[i]:8.3f}\n")

        # make plots
        biasing_force = biasing_force_k[l]
        fig_filename = plot_directory / f"fes-comparison-{l:d}.pdf"
        fig, ax = plt.subplots()
        ax.errorbar(
            bin_center_i, fes_i, yerr=df_i, fmt="x ", elinewidth=0.3, label="MBAR optimal estimate"
        )
        ax.errorbar(
            bin_center_i,
            fes_i_observed,
            yerr=dfes_i_observed,
            fmt="x ",
            elinewidth=0.3,
            label="Observed from single experiment",
        )
        ax.set_title(f"{prefix.title()} - {biasing_force:.2f} pN")
        ax.set_ylabel("Free energy of profile (kT)")
        ax.set_xlabel("Extension (nm)")
        ax.legend()
        fig.savefig(fig_filename)


if __name__ == "__main__":
    main()



"""
todo -- simplify the total energy read in, the kinetic energy read-in, temperature read-in
"""

# =========================================================
# IMPORTS
# =========================================================

from optparse import OptionParser
from pathlib import Path

import numpy as np

import pymbar  # for MBAR analysis
from pymbar import timeseries  # for timeseries analysis


# ===================================================================================================
# INPUT PARAMETERS
# ===================================================================================================
def parse_cli():
    parser = OptionParser()

    parser.add_option(
        "-d",
        "--directory",
        dest="simulation",
        default="energydata",
        help="the directory of the energies we care about",
    )
    parser.add_option(
        "-b",
        "--n_bootstraps",
        dest="n_boots",
        type=int,
        default=0,
        help="Number of bootstrap samples taken",
    )
    parser.add_option(
        "-s",
        "--spacing",
        dest="num_intermediates",
        type="int",
        default=200,
        help="Number of intermediate simulations used to calculate finite differences (default 200)",
    )
    parser.add_option(
        "-f",
        "--finitedifftype",
        dest="dertype",
        default="temperature",
        help='the type of finite difference energy, choice is "temperature" or "beta" [default = %default]',
    )
    parser.add_option(
        "-r",
        "--randomseed",
        dest="rseed",
        type=int,
        default=None,
        help="random seed for bootstraping [default = %default]",
    )
    return parser.parse_args()


options, args = parse_cli()
simulation = options.simulation
n_boots = options.n_boots
num_intermediates = options.num_intermediates
dertype = options.dertype
rseed = options.rseed

# ========================================================
# CONSTANTS
# ========================================================

kB = 0.008314462  # Boltzmann constant (Gas constant) in kJ/(mol*K)
TE_COL_NUM = 11  # The column number of the total energy in ener_box#.output

num_temps = 16  # Last TEMP # + 1 (start counting at 1)
NUM_ITERATIONS = 1000  # The number of energies to be taken and analyzed, starting from the last
# Extra data will be ignored

if dertype == "temperature":  # if the temperatures are equally spaced
    types = "var", "dT", "ddT"
elif dertype == "beta":  # if the inverse temperatures are equally spaced.
    types = "var", "dbeta", "ddbeta"
else:
    sys.exit("type of finite difference not recognized must be 'beta' or 'temperature'")

NTYPES = len(types)

np.random.seed(rseed)  # seed the random numbers

###########################################################
#      For Cv vs T
#                _____
# Cv            /     \                <-- what we expect the graph to look like
#  ____________/       \____________
#                  T
############################################################

# =========================================================
# SUBROUTINES
# =========================================================


def read_total_energies(pathname, colnum):
    """Reads in the TEMP#/ener_box#.output file and parses it, returning an array of energies

    Parameters
    ----------
    pathname : str
        the path to the folder of the simulation
        colnum : int
        column the energy is found in

    Returns
    -------
    array-like
    """
    pathname = Path(pathname)
    print(f"--Reading total energies from {pathname}/...")

    # Initialize Return variables
    E_kn = np.zeros([num_temps, NUM_ITERATIONS])

    # Read files
    for k in range(num_temps):
        # Construct each TEMP#/ener_box#.output name and read in the file
        filename = pathname / f"TEMP{k}" / f"ener_box{k}.output"

        with open(filename) as infile:
            lines = infile.readlines()
            n_lines = len(lines)

        # Initialize arrays for E
        E_from_file = np.zeros(NUM_ITERATIONS)

        # Parse lines in each file
        for n in range(NUM_ITERATIONS):
            # Count down (the 2 is for index purposes(1) and to not use the double-counted last line (1))
            m = n_lines - 2 - n
            elements = lines[m].split()
            E_from_file[n] = float(elements[colnum])

        # Add in the E's for each timestep (n) at this temperature (k)
        E_kn[k] = E_from_file
    return E_kn


def read_simulation_temps(pathname, num_temps):
    """Reads in the various temperatures from each TEMP#/simul.output file by knowing
        beforehand the total number of temperatures (parameter at top)

    Parameters
    ----------
    pathname : str
    num_temps : int

    """
    pathname = Path(pathname)
    print(f"--Reading temperatures from {pathname}/...")

    # Initialize return variable
    temps_from_file = np.zeros(num_temps)

    for k in range(num_temps):
        filename = pathname / f"TEMP{k}" / f"simul{k}.output"
        with open(filename) as f:
            for line in f:
                if line[0:11] == "Temperature":
                    vals = line.split(":")
                    break
            temps_from_file[k] = float(vals[1])

    return temps_from_file


def print_results(string, E, dE, Cv, dCv, types):
    print(string)
    print("Temperature    dA        <E> +/- d<E>  ", end=" ")
    for t in types:
        print(f"    Cv +/- dCv ({t})", end=" ")
    print()
    print(
        "------------------------------------------------------------------------------------------------------"
    )
    for k in range(originalK, K):
        print(
            f"{temp_k[k]:8.3f} {mbar.f_k[k] / beta_k[k]:8.3f} {E[k]:9.3f} +/- {dE[k]:5.3f}",
            end=" ",
        )
        for i in range(len(types)):
            if Cv[k, i, 0] < -100000.0:
                print("         N/A          ", end=" ")
            else:
                print(f"    {Cv[k, i, 0]:7.4f} +/- {dCv[k, i]:6.4f}", end=" ")
        print()


# ========================================================================
# MAIN
# ========================================================================

# ------------------------------------------------------------------------
# Read Data From File
# ------------------------------------------------------------------------

print()
print("Preparing data:")
T_from_file = read_simulation_temps(simulation, num_temps)
E_from_file = read_total_energies(simulation, TE_COL_NUM)
K = len(T_from_file)
N_k = np.zeros(K, dtype=int)
g = np.zeros(K)

for k in range(K):  # subsample the energies
    g[k] = timeseries.statistical_inefficiency(E_from_file[k])
    # indices of uncorrelated samples
    indices = np.array(timeseries.subsample_correlated_data(E_from_file[k], g=g[k]))
    N_k[k] = len(indices)  # number of uncorrelated samples
    E_from_file[k, 0 : N_k[k]] = E_from_file[k, indices]

# ------------------------------------------------------------------------
# Insert Intermediate T's and corresponding blank U's and E's
# ------------------------------------------------------------------------
temp_k = T_from_file
minT = T_from_file[0]
maxT = T_from_file[len(T_from_file) - 1]
# beta = 1/(k*BT)
# T = 1/(kB*beta)
if dertype == "temperature":
    minv = minT
    maxv = maxT
elif dertype == "beta":  # actually going in the opposite direction as beta for logistical reasons
    minv = 1 / (kB * minT)
    maxv = 1 / (kB * maxT)
delta = (maxv - minv) / (num_intermediates - 1)
originalK = len(temp_k)

print("--Adding intermediate temperatures...")

val_k = []
currentv = minv
if dertype == "temperature":
    # Loop, inserting equally spaced T's at which we are interested in the properties
    while currentv <= maxv:
        val_k = np.append(val_k, currentv)
        currentv = currentv + delta
    temp_k = np.concatenate((temp_k, np.array(val_k)))
elif dertype == "beta":
    # Loop, inserting equally spaced T's at which we are interested in the properties
    while currentv >= maxv:
        val_k = np.append(val_k, currentv)
        currentv = currentv + delta
    temp_k = np.concatenate((temp_k, (1 / (kB * np.array(val_k)))))

# Update number of states
K = len(temp_k)
# Loop, inserting E's into blank matrix (leaving blanks only where new Ts are inserted)

# Number of samples (n) for each state (k) = number of iterations/energies
Nall_k = np.zeros([K], dtype=int)
E_kn_files = np.zeros([K, NUM_ITERATIONS])

for k in range(originalK):
    E_kn_files[k, 0 : N_k[k]] = E_from_file[k, 0 : N_k[k]]
    Nall_k[k] = N_k[k]

# ------------------------------------------------------------------------
# Compute inverse temperatures
# ------------------------------------------------------------------------
beta_k = 1 / (kB * temp_k)

# ------------------------------------------------------------------------
# Compute reduced potential energies
# ------------------------------------------------------------------------

print("--Computing reduced energies...")

# u_kln is reduced pot. ener. of segment n of temp k evaluated at temp l
u_kn = np.zeros([K, np.sum(N_k)])
# u_kln is reduced pot. ener. of segment n of temp k evaluated at temp l
E_kn_samp = np.zeros([K, NUM_ITERATIONS])
# we add +1 to the bootstrap number, as the zeroth bootstrap sample is the original
n_boots_work = n_boots + 1

allCv_expect = np.zeros([K, NTYPES, n_boots_work])
dCv_expect = np.zeros([K, NTYPES])
allE_expect = np.zeros([K, n_boots_work])
allE2_expect = np.zeros([K, n_boots_work])
dE_expect = np.zeros([K])


for n in range(n_boots_work):
    if n > 0:
        print(f"Bootstrap: {n:d}/{n_boots:d}")

    for k in range(K):
        # resample the results:
        if Nall_k[k] > 0:
            if n == 0:  # don't randomize the first one
                booti = np.array(range(N_k[k]))
            else:
                booti = np.random.randint(Nall_k[k], size=Nall_k[k])
            E_kn_samp[k, 0 : Nall_k[k]] = E_kn_files[k, booti]

    for l in range(K):
        nsum = 0
        for k in range(K):
            u_kn[l, nsum : nsum + Nall_k[k]] = beta_k[l] * E_kn_samp[k, 0 : Nall_k[k]]
            nsum = nsum + Nall_k[k]
    # ------------------------------------------------------------------------
    # Initialize MBAR
    # ------------------------------------------------------------------------

    # Initialize MBAR with Newton-Raphson
    if n == 0:  # only print this information the first time
        print()
        print("Initializing MBAR:")
        print(f"--K = number of Temperatures with data = {originalK:d}")
        print(f"--L = number of total Temperatures = {K:d}")
        print(f"--N = number of Energies per Temperature = {np.max(Nall_k):d}")

    if n == 0:
        initial_f_k = None  # start from zero
    else:
        initial_f_k = mbar.f_k  # start from the previous final free energies to speed convergence

    mbar = pymbar.MBAR(
        u_kn,
        Nall_k,
        verbose=False,
        relative_tolerance=1e-12,
        solver_protocol="robust",
        initial_f_k=initial_f_k,
    )

    # ------------------------------------------------------------------------
    # Compute Expectations for E_kt and E2_kt as E_expect and E2_expect
    # ------------------------------------------------------------------------

    print("")
    print("Computing Expectations for E...")
    E_kn = u_kn.copy()
    for k in range(K):
        # get the 'unreduced' potential -- we can't take differences of reduced potentials
        # because the beta is different; math is much more confusing with derivatives of the reduced potentials.
        E_kn[k, :] *= beta_k[k] ** (-1)
    results = mbar.compute_expectations(E_kn, state_dependent=True)
    E_expect = results["mu"]
    dE_expect = results["sigma"]
    allE_expect[:, n] = E_expect[:]

    # expectations for the differences, which we need for numerical derivatives
    results = mbar.compute_expectations(E_kn, output="differences", state_dependent=True)
    DeltaE_expect = results["mu"]
    dDeltaE_expect = results["sigma"]
    print("Computing Expectations for E^2...")

    results = mbar.compute_expectations(E_kn**2, state_dependent=True)
    E2_expect = results["mu"]
    dE2_expect = results["sigma"]
    allE2_expect[:, n] = E2_expect[:]

    results = mbar.compute_free_energy_differences()
    df_ij = results["Delta_f"]
    ddf_ij = results["dDelta_f"]

    # ------------------------------------------------------------------------
    # Compute Cv for NVT simulations as <E^2> - <E>^2 / (RT^2)
    # ------------------------------------------------------------------------

    if n == 0:
        print()
        print("Computing Heat Capacity as ( <E^2> - <E>^2 ) / ( R*T^2 ) and as d<E>/dT")

    # Problem is that we don't have a good uncertainty estimate for the variance.
    # Try a silly trick: but it doesn't work super well.
    # An estimator of the variance of the standard estimator of th evariance is
    # var(sigma^2) = (sigma^4)*[2/(n-1)+kurt/n]. If we assume the kurtosis is low
    # (which it will be for sufficiently many samples), then we can say that
    # d(sigma^2) = sigma^2 sqrt[2/(n-1)].
    # However, dE_expect**2 is already an estimator of sigma^2/(n-1)
    # Cv = sigma^2/kT^2, so d(Cv) = d(sigma^2)/kT^2 = sigma^2*[sqrt(2/(n-1)]/kT^2
    # we just need an estimate of n-1, but we can try to get that by var(dE)/dE_expect**2
    # it's within 50% or so, but that's not good enough.

    allCv_expect[:, 0, n] = (E2_expect - (E_expect * E_expect)) / (kB * temp_k**2)

    ####################################
    # C_v by fluctuation formula
    ####################################

    # Cv = (A - B^2) / (kT^2)
    # d2(Cv) = [1/(kT^2)]^2 [(dCv/dA)^2*d2A + 2*dCv*(dCv/dA)*(dCv/dB)*dAdB + (dCv/dB)^2*d2B]
    # = [1/(kT^2)]^2 [d2A - 4*B*dAdB + 4*B^2*d2B]
    # But this formula is not working for uncertainies!

    if n == 0:
        # sigma^2 / (sigma^2/n) = effective number of samples
        N_eff = (E2_expect - (E_expect * E_expect)) / dE_expect**2
        dCv_expect[:, 0] = allCv_expect[:, 0, n] * np.sqrt(2 / N_eff)

    # only loop over the points that will be plotted, not the ones that
    for i in range(originalK, K):
        # Now, calculae heat capacity by T-differences
        im = i - 1
        ip = i + 1
        if i == originalK:
            im = originalK
        if i == K - 1:
            ip = i

        ####################################
        # C_v by first derivative of energy
        ####################################

        if dertype == "temperature":  # temperature derivative
            # C_v = d<E>/dT
            allCv_expect[i, 1, n] = (DeltaE_expect[im, ip]) / (temp_k[ip] - temp_k[im])
            if n == 0:
                dCv_expect[i, 1] = (dDeltaE_expect[im, ip]) / (temp_k[ip] - temp_k[im])
        elif dertype == "beta":  # beta derivative
            # Cv = d<E>/dT = dbeta/dT d<E>/beta = - kB*T(-2) d<E>/dbeta  = - kB beta^2 d<E>/dbeta
            allCv_expect[i, 1, n] = (
                kB * beta_k[i] ** 2 * (DeltaE_expect[ip, im]) / (beta_k[ip] - beta_k[im])
            )
            if n == 0:
                dCv_expect[i, 1] = (
                    -kB * beta_k[i] ** 2 * (dDeltaE_expect[ip, im]) / (beta_k[ip] - beta_k[im])
                )

        #########################################
        # C_v by second derivative of free energy
        #########################################

        if dertype == "temperature":
            # C_v = d<E>/dT = d/dT k_B T^2 df/dT = 2*T*df/dT + T^2*d^2f/dT^2

            if (i == originalK) or (i == K - 1):
                # We can't calculate this, set a number that will be printed as NAN
                allCv_expect[i, 2, n] = -10000000.0
            else:
                allCv_expect[i, 2, n] = (
                    kB
                    * temp_k[i]
                    * (
                        2 * df_ij[ip, im] / (temp_k[ip] - temp_k[im])
                        + temp_k[i]
                        * (df_ij[ip, i] - df_ij[i, im])
                        / ((temp_k[ip] - temp_k[im]) / (ip - im)) ** 2
                    )
                )

            if n == 0:
                # Previous work to calculate the uncertainty commented out, should be cleaned up eventually
                # all_Cv_expect[i,2,n] = kB*temp_k[i]*(2*df_ij[ip,i]+df_ij[i,im]/(temp_k[ip]-temp_k[im]) + temp_k[i]*(df_ij[ip,i]-df_ij[i,im])/(temp_k[ip]-temp_k[i])**2)
                # all_Cv_expect[i,2,n] = kB*([2*temp_k[i]/(temp_k[ip]-temp_k[im]) + temp_k[i]**2/(temp_k[ip]-temp_k[i])**2]*df_ij[ip,i] + [2*temp_k[i]/(temp_k[ip]-temp_k[im]) - temp_k[i]**2/(temp_k[ip]-temp_k[i])**2]) df_ij[i,im]
                # all_Cv_expect[i,2,n] = kB*(A df_ij[ip,i] + B df_ij[i,im]
                A = (
                    2 * temp_k[i] / (temp_k[ip] - temp_k[im])
                    + 4 * temp_k[i] ** 2 / (temp_k[ip] - temp_k[im]) ** 2
                )
                B = (
                    2 * temp_k[i] / (temp_k[ip] - temp_k[im])
                    + 4 * temp_k[i] ** 2 / (temp_k[ip] - temp_k[im]) ** 2
                )
                # dCv_expect[i,2,n] = kB* [(A ddf_ij[ip,i])**2 + (B sdf_ij[i,im])**2 + 2*A*B*cov(df_ij[ip,i],df_ij[i,im])
                # This isn't it either: need to figure out that last term.
                dCv_expect[i, 2] = kB * ((A * ddf_ij[ip, i]) ** 2 + (B * ddf_ij[i, im]) ** 2)
                # Would need to add function computing covariance of DDG, (A-B)-(C-D)

        elif dertype == "beta":
            # if beta is evenly spaced, rather than t, we can do 2nd derivative in beta
            # C_v = d<E>/dT = d/dT (df/dbeta) = dbeta/dT d/dbeta (df/dbeta) = -k_b beta^2 df^2/d^2beta
            if (i == originalK) or (i == K - 1):
                # Flag as N/A -- we don't try to compute at the endpoints for now
                allCv_expect[i, 2, n] = -10000000.0
            else:
                allCv_expect[i, 2, n] = (
                    kB
                    * beta_k[i] ** 2
                    * (df_ij[ip, i] - df_ij[i, im])
                    / ((beta_k[ip] - beta_k[im]) / (ip - im)) ** 2
                )
            if n == 0:
                dCv_expect[i, 2] = (
                    kB
                    * (beta_k[i]) ** 2
                    * (ddf_ij[ip, i] - ddf_ij[i, im])
                    / ((beta_k[ip] - beta_k[im]) / (ip - im)) ** 2
                )
                # also wrong, need to be fixed.

    if n == 0:
        print(
            "WARNING: only the first derivative (dT) analytic error estimates can currently be trusted."
        )
        print("They are the only ones reasonably close to bootstrap, within 10-15% at all T.")
        print()
        print_results(
            "Analytic Error Estimates", E_expect, dE_expect, allCv_expect, dCv_expect, types
        )

if n_boots > 0:
    Cv_boot = np.zeros([K, NTYPES])
    dCv_boot = np.zeros([K, NTYPES])
    dE_boot = np.zeros([K])

    for k in range(K):
        for i in range(NTYPES):
            # for these averages, don't include the first one, because it's the non-bootstrapped one.
            Cv_boot[k, i] = np.mean(allCv_expect[k, i, 1:n_boots_work])
            dCv_boot[k, i] = np.std(allCv_expect[k, i, 1:n_boots_work])
            dE_boot[k] = np.std(allE_expect[k, 1:n_boots_work])
    print_results(
        "Bootstrap Error Estimates", allE_expect[:, 0], dE_boot, allCv_expect, dCv_boot, types
    )
